@@ -53,6 +53,8 @@
 #define CFG_WRITE_PROXY_V2 "write-proxy-v2"
 #define CFG_PEM_FILE "pem-file"
 #define CFG_PROXY_PROXY "proxy-proxy"
+#define CFG_ALPN_PROTOS "alpn-protos"
+#define CFG_PARAM_ALPN_PROTOS 48173
 #define CFG_BACKEND_CONNECT_TIMEOUT "backend-connect-timeout"
 #define CFG_SSL_HANDSHAKE_TIMEOUT "ssl-handshake-timeout"
 #define CFG_RECV_BUFSIZE "recv-bufsize"
@@ -62,6 +64,8 @@
 #define CFG_RING_DATA_LEN "ring-data-len"
 #define CFG_PIDFILE "pidfile"
 #define CFG_SNI_NOMATCH_ABORT "sni-nomatch-abort"
+#define CFG_OCSP_DIR "ocsp-dir"
+#define CFG_TLS_PROTOS "tls-protos"
 
 #ifdef USE_SHARED_CACHE
 	#define CFG_SHARED_CACHE "shared-cache"
@@ -82,11 +86,17 @@
 extern FILE *yyin;
 extern int yyparse(hitch_config *);
 
+void cfg_cert_file_free(struct cfg_cert_file **cfptr);
+
 // END: configuration parameters
 
 static char error_buf[CONFIG_BUF_SIZE];
 static char tmp_buf[150];
 
+/* declare static printf like functions: */
+static void config_error_set(char *fmt, ...)
+	__attribute__((format(printf, 1, 2)));
+
 static void
 config_error_set(char *fmt, ...)
 {
@@ -110,6 +120,39 @@ config_error_get(void)
 	return error_buf;
 }
 
+struct front_arg *
+front_arg_new(void)
+{
+	struct front_arg *fa;
+
+	ALLOC_OBJ(fa, FRONT_ARG_MAGIC);
+	AN(fa);
+	fa->match_global_certs = -1;
+	fa->sni_nomatch_abort = -1;
+	fa->selected_protos = 0;
+	fa->prefer_server_ciphers = -1;
+
+	return (fa);
+}
+
+void
+front_arg_destroy(struct front_arg *fa)
+{
+	struct cfg_cert_file *cf, *cftmp;
+
+	CHECK_OBJ_NOTNULL(fa, FRONT_ARG_MAGIC);
+	free(fa->ip);
+	free(fa->port);
+	free(fa->pspec);
+	free(fa->ciphers);
+	HASH_ITER(hh, fa->certs, cf, cftmp) {
+		CHECK_OBJ_NOTNULL(cf, CFG_CERT_FILE_MAGIC);
+		HASH_DEL(fa->certs, cf);
+		cfg_cert_file_free(&cf);
+	}
+	FREE_OBJ(fa);
+}
+
 hitch_config *
 config_new(void)
 {
@@ -121,12 +164,15 @@ config_new(void)
 
 	// set default values
 
-	r->ETYPE              = ENC_TLS;
 	r->PMODE              = SSL_SERVER;
+	r->SELECTED_TLS_PROTOS= 0;
 	r->WRITE_IP_OCTET     = 0;
 	r->WRITE_PROXY_LINE_V1= 0;
 	r->WRITE_PROXY_LINE_V2= 0;
 	r->PROXY_PROXY_LINE   = 0;
+	r->ALPN_PROTOS        = NULL;
+	r->ALPN_PROTOS_LV     = NULL;
+	r->ALPN_PROTOS_LV_LEN = 0;
 	r->CHROOT             = NULL;
 	r->UID                = -1;
 	r->GID                = -1;
@@ -137,14 +183,17 @@ config_new(void)
 	r->ENGINE             = NULL;
 	r->BACKLOG            = 100;
 	r->SNI_NOMATCH_ABORT  = 0;
-	r->CERT_DEFAULT = NULL;
-	r->CERT_FILES = NULL;
-	r->LISTEN_ARGS = NULL;
-	ALLOC_OBJ(fa, FRONT_ARG_MAGIC);
+	r->CERT_DEFAULT	      = NULL;
+	r->CERT_FILES         = NULL;
+	r->LISTEN_ARGS        = NULL;
+	fa = front_arg_new();
 	fa->port = strdup("8443");
 	fa->pspec = strdup("default");
 	HASH_ADD_KEYPTR(hh, r->LISTEN_ARGS, fa->pspec, strlen(fa->pspec), fa);
 	r->LISTEN_DEFAULT = fa;
+	r->OCSP_VFY = 0;
+	r->OCSP_RESP_TMO = 10.0;
+	r->OCSP_CONN_TMO = 4.0;
 
 #ifdef USE_SHARED_CACHE
 	r->SHARED_CACHE       = 0;
@@ -195,34 +244,26 @@ config_destroy(hitch_config *cfg)
 	HASH_ITER(hh, cfg->LISTEN_ARGS, fa, ftmp) {
 		CHECK_OBJ_NOTNULL(fa, FRONT_ARG_MAGIC);
 		HASH_DEL(cfg->LISTEN_ARGS, fa);
-		free(fa->ip);
-		free(fa->port);
-		free(fa->pspec);
-		free(fa->ciphers);
-		HASH_ITER(hh, fa->certs, cf, cftmp) {
-			CHECK_OBJ_NOTNULL(cf, CFG_CERT_FILE_MAGIC);
-			HASH_DEL(fa->certs, cf);
-			free(cf->filename);
-			FREE_OBJ(cf);
-		}
-		FREE_OBJ(fa);
+		front_arg_destroy(fa);
 	}
 	free(cfg->BACK_IP);
 	free(cfg->BACK_PORT);
 	HASH_ITER(hh, cfg->CERT_FILES, cf, cftmp) {
 		CHECK_OBJ_NOTNULL(cf, CFG_CERT_FILE_MAGIC);
 		HASH_DEL(cfg->CERT_FILES, cf);
-		free(cf->filename);
-		FREE_OBJ(cf);
+		cfg_cert_file_free(&cf);
 	}
 
 	if (cfg->CERT_DEFAULT != NULL) {
-		free(cfg->CERT_DEFAULT->filename);
-		FREE_OBJ(cfg->CERT_DEFAULT);
+		cfg_cert_file_free(&cfg->CERT_DEFAULT);
 	}
+
 	free(cfg->CIPHER_SUITE);
 	free(cfg->ENGINE);
 	free(cfg->PIDFILE);
+	free(cfg->OCSP_DIR);
+	free(cfg->ALPN_PROTOS);
+	free(cfg->ALPN_PROTOS_LV);
 
 #ifdef USE_SHARED_CACHE
 	free(cfg->SHCUPD_IP);
@@ -397,44 +438,96 @@ config_param_val_long(char *str, long *dst, int positive_only)
 	return 1;
 }
 
+static double
+mtim2double(const struct stat *sb)
+{
+	double d = sb->st_mtime;
+
+#if defined(HAVE_STRUCT_STAT_ST_MTIM)
+	d += sb->st_mtim.tv_nsec * 1e-9;
+#elif defined(HAVE_STRUCT_STAT_ST_MTIMESPEC)
+	d += sb->st_mtimespec.tv_nsec * 1e-9;
+#endif
+	return (d);
+}
+
+struct cfg_cert_file *
+cfg_cert_file_new(void)
+{
+	struct cfg_cert_file *cert;
+	ALLOC_OBJ(cert, CFG_CERT_FILE_MAGIC);
+	AN(cert);
+	cert->ocsp_vfy = -1;
+	return (cert);
+}
+
+void
+cfg_cert_file_free(struct cfg_cert_file **cfptr)
+{
+	struct cfg_cert_file *cf;
+
+	CHECK_OBJ_NOTNULL(*cfptr, CFG_CERT_FILE_MAGIC);
+	cf = *cfptr;
+	free(cf->filename);
+	free(cf->ocspfn);
+	FREE_OBJ(cf);
+	*cfptr = NULL;
+}
+
 int
-config_param_pem_file(char *filename, struct cfg_cert_file **cfptr)
+cfg_cert_vfy(struct cfg_cert_file *cf)
 {
 	struct stat st;
-	struct cfg_cert_file *cert;
 
-	*cfptr = NULL;
+	CHECK_OBJ_NOTNULL(cf, CFG_CERT_FILE_MAGIC);
 
-	if (strlen(filename) <= 0)
+	AN(cf->filename);
+
+	if (cf->filename == NULL || strlen(cf->filename) <= 0)
 		return (0);
 
-	if (stat(filename, &st) != 0) {
+	if (stat(cf->filename, &st) != 0) {
 		config_error_set("Unable to stat x509 "
-		    "certificate PEM file '%s': ", filename,
+		    "certificate PEM file '%s': %s", cf->filename,
 		    strerror(errno));
 		return (0);
 	}
-	if (! S_ISREG(st.st_mode)) {
+	if (!S_ISREG(st.st_mode)) {
 		config_error_set("Invalid x509 certificate "
-		    "PEM file '%s': Not a file.", filename);
+		    "PEM file '%s': Not a file.", cf->filename);
 		return (0);
 	}
+	cf->mtim = mtim2double(&st);
 
-	ALLOC_OBJ(cert, CFG_CERT_FILE_MAGIC);
-	AN(cert);
-	config_assign_str(&cert->filename, filename);
-	cert->mtim = st.st_mtime;
-#if defined(HAVE_STRUCT_STAT_ST_MTIM)
-	cert->mtim += st.st_mtim.tv_nsec * 1e-9;
-#elif defined(HAVE_STRUCT_STAT_ST_MTIMESPEC)
-	cert->mtim += st.st_mtimespec.tv_nsec * 1e-9;
-#endif
+	if (cf->ocspfn != NULL) {
+		if (stat(cf->ocspfn, &st) == -1) {
+			config_error_set("Unable to stat OCSP "
+			    "stapling file '%s': %s", cf->ocspfn,
+			    strerror(errno));
+			return (0);
+		}
+		if (!S_ISREG(st.st_mode)) {
+			config_error_set("Invalid OCSP stapling file "
+			    "'%s': Not a file.", cf->ocspfn);
+			return (0);
+		}
+		cf->ocsp_mtim = mtim2double(&st);
+	}
 
-
-	*cfptr = cert;
 	return (1);
-
 }
+
+void
+cfg_cert_add(struct cfg_cert_file *cf, struct cfg_cert_file **dst)
+{
+	CHECK_OBJ_NOTNULL(cf, CFG_CERT_FILE_MAGIC);
+	AN(dst);
+	CHECK_OBJ_ORNULL(*dst, CFG_CERT_FILE_MAGIC);
+
+	HASH_ADD_KEYPTR(hh, *dst, cf->filename,
+	    strlen(cf->filename), cf);
+}
+
 #ifdef USE_SHARED_CACHE
 /* Parse mcast and ttl options */
 int
@@ -453,7 +546,7 @@ config_param_shcupd_mcastif(char *str, char **iface, char **ttl)
 		if (!strcmp(str, "*"))
 			*iface = NULL;
 		else
-			*iface = str;
+			*iface = strdup(str);
 		*ttl = NULL;
 		return 1;
 	}
@@ -462,9 +555,9 @@ config_param_shcupd_mcastif(char *str, char **iface, char **ttl)
 	}
 	else {
 		*sp = 0;
-		*iface = str;
+		*iface = strdup(str);
 	}
-	*ttl = sp + 1;
+	*ttl = strdup(sp + 1);
 
 	return 1;
 }
@@ -541,29 +634,12 @@ config_param_shcupd_peer(char *str, hitch_config *cfg)
 
 #endif /* USE_SHARED_CACHE */
 
-struct front_arg *
-front_arg_new(void)
-{
-	struct front_arg *fa;
-
-	ALLOC_OBJ(fa, FRONT_ARG_MAGIC);
-	AN(fa);
-	fa->match_global_certs = -1;
-	fa->sni_nomatch_abort = -1;
-	fa->etype = ENC_TLS;
-	fa->prefer_server_ciphers = -1;
-
-	return (fa);
-}
-
 int
 front_arg_add(hitch_config *cfg, struct front_arg *fa)
 {
 	struct vsb pspec;
 
 	CHECK_OBJ_NOTNULL(fa, FRONT_ARG_MAGIC);
-	AN(fa->port);
-
 	if (cfg->LISTEN_DEFAULT != NULL) {
 		/* drop default listen arg. */
 		struct front_arg *def = NULL;
@@ -578,10 +654,16 @@ front_arg_add(hitch_config *cfg, struct front_arg *fa)
 	}
 
 	VSB_new(&pspec, NULL, 0, VSB_AUTOEXTEND);
-	VSB_printf(&pspec, "[%s]:", fa->ip);
-	VSB_cat(&pspec, fa->port);
+	VSB_printf(&pspec, "[%s]:%s", fa->ip, fa->port);
 	VSB_finish(&pspec);
 	fa->pspec = VSB_data(&pspec);
+
+	if (fa->port == NULL) {
+		config_error_set("No port number specified "
+		    "for frontend '%s'", fa->pspec);
+		return (0);
+	}
+
 	HASH_ADD_KEYPTR(hh, cfg->LISTEN_ARGS, fa->pspec,
 	    strlen(fa->pspec), fa);
 
@@ -614,9 +696,9 @@ config_param_validate(char *k, char *v, hitch_config *cfg,
 	assert(strlen(k) >= 2);
 
 	if (strcmp(k, "tls") == 0) {
-		cfg->ETYPE = ENC_TLS;
+		cfg->SELECTED_TLS_PROTOS = TLS_OPTION_PROTOS;
 	} else if (strcmp(k, "ssl") == 0) {
-		cfg->ETYPE = ENC_SSL;
+		cfg->SELECTED_TLS_PROTOS = SSL_OPTION_PROTOS;
 	} else if (strcmp(k, CFG_CIPHERS) == 0) {
 		if (strlen(v) > 0) {
 			config_assign_str(&cfg->CIPHER_SUITE, v);
@@ -637,16 +719,19 @@ config_param_validate(char *k, char *v, hitch_config *cfg,
 		    &fa->ip, &fa->port, &certfile, 1);
 		if (r != 0) {
 			if (certfile != NULL) {
-				r = config_param_pem_file(certfile, &cert);
-				if (r != 0) {
-					AN(cert);
-					HASH_ADD_KEYPTR(hh,
-					    fa->certs, cert->filename,
-					    strlen(cert->filename), cert);
-				}
+				cert = cfg_cert_file_new();
+				config_assign_str(&cert->filename, certfile);
+				r = cfg_cert_vfy(cert);
+				if (r != 0)
+					cfg_cert_add(cert, &fa->certs);
+				else
+					cfg_cert_file_free(&cert);
 				free(certfile);
-		}
-			r = front_arg_add(cfg, fa);
+			}
+			if (r != 0)
+				r = front_arg_add(cfg, fa);
+			else
+				FREE_OBJ(fa);
 		}
 	} else if (strcmp(k, CFG_BACKEND) == 0) {
 		free(cfg->BACK_PORT);
@@ -683,8 +768,7 @@ config_param_validate(char *k, char *v, hitch_config *cfg,
 			} else {
 				if (! S_ISDIR(st.st_mode)) {
 					config_error_set("Bad chroot directory "
-					    "'%s': Not a directory.", v,
-					    strerror(errno));
+					    "'%s': Not a directory", v);
 					r = 0;
 				} else {
 					config_assign_str(&cfg->CHROOT, v);
@@ -744,19 +828,23 @@ config_param_validate(char *k, char *v, hitch_config *cfg,
 		r = config_param_val_bool(v, &cfg->WRITE_PROXY_LINE_V2);
 	} else if (strcmp(k, CFG_PROXY_PROXY) == 0) {
 		r = config_param_val_bool(v, &cfg->PROXY_PROXY_LINE);
+	} else if (strcmp(k, CFG_ALPN_PROTOS) == 0) {
+		if (strlen(v) > 0) {
+			config_assign_str(&cfg->ALPN_PROTOS, v);
+		}
 	} else if (strcmp(k, CFG_PEM_FILE) == 0) {
 		struct cfg_cert_file *cert;
-		r = config_param_pem_file(v, &cert);
+		cert = cfg_cert_file_new();
+		config_assign_str(&cert->filename, v);
+		r = cfg_cert_vfy(cert);
 		if (r != 0) {
-			AN(cert);
 			if (cfg->CERT_DEFAULT != NULL) {
 				struct cfg_cert_file *tmp = cfg->CERT_DEFAULT;
-				HASH_ADD_KEYPTR(hh, cfg->CERT_FILES,
-				    tmp->filename, strlen(tmp->filename),
-				    tmp);
+				cfg_cert_add(tmp, &cfg->CERT_FILES);
 			}
 			cfg->CERT_DEFAULT = cert;
-		}
+		} else
+			cfg_cert_file_free(&cert);
 	} else if (strcmp(k, CFG_BACKEND_CONNECT_TIMEOUT) == 0) {
 		r = config_param_val_int(v, &cfg->BACKEND_CONNECT_TIMEOUT, 1);
 	} else if (strcmp(k, CFG_SSL_HANDSHAKE_TIMEOUT) == 0) {
@@ -779,6 +867,8 @@ config_param_validate(char *k, char *v, hitch_config *cfg,
 		r = config_param_val_int(v, &cfg->RING_DATA_LEN, 1);
 	} else if (strcmp(k, CFG_SNI_NOMATCH_ABORT) == 0) {
 		r = config_param_val_bool(v, &cfg->SNI_NOMATCH_ABORT);
+	} else if (strcmp(k, CFG_OCSP_DIR) == 0) {
+		config_assign_str(&cfg->OCSP_DIR, v);
 	} else {
 		fprintf(
 			stderr,
@@ -989,6 +1079,9 @@ config_print_usage_fd(char *prog, FILE *out)
 			"submits an unrecognized SNI server name\n" );
 	fprintf(out, "                             (Default: %s)\n",
 			config_disp_bool(cfg->SNI_NOMATCH_ABORT));
+	fprintf(out, "      --ocsp-dir=DIR         Set OCSP staple cache directory\n");
+	fprintf(out, "                             This enables automated retrieval and stapling of OCSP responses\n");
+	fprintf(out, "                             (Default: \"%s\")\n", config_disp_str(cfg->OCSP_DIR));
 	fprintf(out, "\n");
 	fprintf(out, "  -t  --test                 Test configuration and exit\n");
 	fprintf(out, "  -p  --pidfile=FILE         PID file\n");
@@ -1047,12 +1140,15 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 		{ CFG_WRITE_PROXY_V2, 0, &cfg->WRITE_PROXY_LINE_V2, 1 },
 		{ CFG_WRITE_PROXY, 0, &cfg->WRITE_PROXY_LINE_V2, 1 },
 		{ CFG_PROXY_PROXY, 0, &cfg->PROXY_PROXY_LINE, 1 },
+		{ CFG_ALPN_PROTOS, 1, NULL, CFG_PARAM_ALPN_PROTOS },
 		{ CFG_SNI_NOMATCH_ABORT, 0, &cfg->SNI_NOMATCH_ABORT, 1 },
+		{ CFG_OCSP_DIR, 1, NULL, 'o' },
 		{ "test", 0, NULL, 't' },
 		{ "version", 0, NULL, 'V' },
 		{ "help", 0, NULL, 'h' },
 		{ 0, 0, 0, 0 }
 	};
+#define SHORT_OPTS "c:e:Ob:f:n:B:C:U:p:P:M:k:r:u:g:qstVho:"
 
 	if (argc == 1) {
 		config_print_usage(argv[0]);
@@ -1060,11 +1156,30 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 		return (1);
 	}
 
+	/* First do a pass over the args string to see if there was a
+	 * config file present. If so, apply its options first in
+	 * order to let them be overridden by the command line.  */
+	while (1) {
+		int option_index = 0;
+		c = getopt_long(argc, argv, SHORT_OPTS,
+			long_options, &option_index);
+		if (c == -1)
+			break;
+		else if (c == CFG_PARAM_CFGFILE) {
+			if (config_file_parse(optarg, cfg) != 0) {
+				*retval = 1;
+				return (1);
+			}
+		}
+	}
+
+	int tls_protos_config_file = cfg->SELECTED_TLS_PROTOS;
+
+	optind = 1;
 	while (1) {
 		int ret = 0;
 		int option_index = 0;
-		c = getopt_long(argc, argv,
-			"c:e:Ob:f:n:B:C:U:p:P:M:k:r:u:g:qstVh",
+		c = getopt_long(argc, argv, SHORT_OPTS,
 			long_options, &option_index);
 
 		if (c == -1)
@@ -1074,10 +1189,7 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 		case 0:
 			break;
 		case CFG_PARAM_CFGFILE:
-			if (config_file_parse(optarg, cfg) != 0) {
-				*retval = 1;
-				return (1);
-			}
+			/* Handled above */
 			break;
 		case CFG_PARAM_SYSLOG_FACILITY:
 			ret = config_param_validate(CFG_SYSLOG_FACILITY, optarg, cfg, NULL, 0);
@@ -1151,6 +1263,12 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 			*retval = 0;
 			return (1);
 			break;
+		case 'o':
+			ret = config_param_validate(CFG_OCSP_DIR, optarg, cfg, NULL, 0);
+			break;
+		case CFG_PARAM_ALPN_PROTOS:
+			ret = config_param_validate(CFG_ALPN_PROTOS, optarg, cfg, NULL, 0);
+			break;
 
 		default:
 			config_error_set("Invalid command line parameters. "
@@ -1166,6 +1284,12 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 		}
 	}
 
+	if ((tls || ssl) && tls_protos_config_file != 0) {
+		config_error_set("Deprecated options --tls and --ssl cannot be"
+		    " used to override tls-protos in a config file.");
+		*retval = 1;
+		return (1);
+	}
 	if (tls && ssl) {
 		config_error_set("Options --tls and --ssl are mutually"
 		    " exclusive.");
@@ -1173,10 +1297,12 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 		return (1);
 	} else {
 		if (ssl)
-			cfg->ETYPE = ENC_SSL;
+			cfg->SELECTED_TLS_PROTOS = SSL_OPTION_PROTOS;
 		else if (tls)
-			cfg->ETYPE = ENC_TLS;
+			cfg->SELECTED_TLS_PROTOS = TLS_OPTION_PROTOS;
 	}
+	if (cfg->SELECTED_TLS_PROTOS == 0)
+		cfg->SELECTED_TLS_PROTOS = DEFAULT_TLS_PROTOS;
 
 	if (client)
 		cfg->PMODE = SSL_CLIENT;
@@ -1205,6 +1331,71 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 	}
 #endif
 
+	/* ALPN/NPN protocol negotiation additional configuration and error
+	   handling */
+	if (cfg->ALPN_PROTOS != NULL) {
+		char *error;
+		if (!create_alpn_callback_data(cfg, &error)) {
+			if (error)
+				config_error_set("alpn-protos configuration"
+				    " \"%s\" is bad. %s",
+				    cfg->ALPN_PROTOS, error);
+			else
+				config_error_set("alpn-protos configuration"
+				    " \"%s\" is bad. See man page for more"
+				    " info.",
+				    cfg->ALPN_PROTOS);
+			*retval = 1;
+			return (1);
+		}
+		AN(cfg->ALPN_PROTOS_LV);
+		int multi_proto =
+		    cfg->ALPN_PROTOS_LV[0] != cfg->ALPN_PROTOS_LV_LEN - 1;
+		if (0 == strncmp((char *)cfg->ALPN_PROTOS_LV, "\x8http/1.1", 9)
+		    && !multi_proto)
+			fprintf(stderr, "standard protocol in use\n");
+		if (multi_proto && !cfg->WRITE_PROXY_LINE_V2) {
+			config_error_set("alpn-protos is specified with"
+			    " more than one protocol while proxy-v2 is "
+			    " not selected. This is a configuration"
+			    " error.");
+			*retval = 1;
+			return (1);
+			/* Note that this test was carried out indepenently of
+			   the availability of ALPN / NPN */
+		}
+#if defined(OPENSSL_WITH_NPN) || defined(OPENSSL_WITH_ALPN)
+		/*
+		if (cfg->WRITE_PROXY_LINE_V2)
+			fprintf(stderr, ALPN_NPN_PREFIX_STR
+			    " Negotiated protocol will be communicated to the"
+			    " backend.\n");
+		*/
+#ifndef OPENSSL_WITH_ALPN
+		fprintf(stderr, ALPN_NPN_PREFIX_STR " Warning: Hitch has been"
+		    " compiled against a version of OpenSSL without ALPN"
+		    " support.\n");
+#endif
+#else
+		/* No support for ALPN / NPN support in OpenSSL */
+		if (multi_proto ||
+		    0 != strncmp(cfg->ALPN_PROTOS_LV, "\x8http/1.1", 9)) {
+			config_error_set("This is compiled against SSL version"
+			    " %lx, which does not have no NPN or ALPN support,"
+			    " yet alpn-protos has been set to %s.",
+			    OPENSSL_VERSION_NUMBER, cfg->ALPN_PROTOS);
+			*retval = 1;
+			return (1);
+		}
+		else
+			fprintf(stderr, "This is compiled against SSL version"
+			    " %lx, which does not have no NPN or ALPN support,"
+			    " but since alpn-protos has been set to http/1.1,"
+			    " we carry on without ALPN/NPN.\n",
+			    OPENSSL_VERSION_NUMBER);
+#endif
+	}
+
 	// Any arguments left are presumed to be PEM files
 	argc -= optind;
 	argv += optind;
@@ -1215,7 +1406,6 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 		}
 	}
 
-
 	if (cfg->PMODE == SSL_SERVER && cfg->CERT_DEFAULT == NULL) {
 		struct front_arg *fa, *fatmp;
 		HASH_ITER(hh, cfg->LISTEN_ARGS, fa, fatmp) {
@@ -1229,5 +1419,72 @@ config_parse_cli(int argc, char **argv, hitch_config *cfg, int *retval)
 		}
 	}
 
+	if (cfg->OCSP_DIR != NULL) {
+		struct stat sb;
+		if (stat(cfg->OCSP_DIR, &sb) != 0) {
+			config_error_set("ocsp-dir: Unable to stat directory"
+			    " '%s': %s'.", cfg->OCSP_DIR, strerror(errno));
+			*retval = 1;
+			return (1);
+		} else {
+			if (!S_ISDIR(sb.st_mode)) {
+				config_error_set("Bad ocsp-dir "
+				    "'%s': Not a directory.", cfg->OCSP_DIR);
+				*retval = 1;
+				return (1);
+			}
+		}
+	}
+
 	return (0);
 }
+
+int create_alpn_callback_data(hitch_config *cfg, char **error)
+{
+	size_t i = 1, j, l;
+
+	AN(cfg->ALPN_PROTOS);
+	l = strlen(cfg->ALPN_PROTOS);
+	cfg->ALPN_PROTOS_LV = malloc(l + 1);
+	AN(cfg->ALPN_PROTOS_LV);
+
+	// first remove spaces while copying to cfg->ALPN_PROTOS_LV
+	for(j = 0; j < l; j++) {
+		if (!isspace(cfg->ALPN_PROTOS[j])) {
+			cfg->ALPN_PROTOS_LV[i] = cfg->ALPN_PROTOS[j];
+			i++;
+		}
+	}
+	l = i - 1; // same as before iff cfg->ALPN_PROTOS has no spaces
+	i = 0; // location of next "length" byte
+	for(j = 1; j <= l; j++) {
+		if (cfg->ALPN_PROTOS_LV[j] == ',') {
+			if (i + 1 == j) {
+				*error = "alpn-protos has empty proto in list";
+				return (0); // failure
+			}
+			if (j - i > 256) {
+				free(cfg->ALPN_PROTOS_LV);
+				cfg->ALPN_PROTOS_LV = NULL;
+				*error = "alpn protocol too long";
+				return (0);
+			}
+			cfg->ALPN_PROTOS_LV[i] = (unsigned char)(j - i - 1);
+			i = j;
+		}
+	}
+	if (i == j) {
+		// alpn-protos ends with a comma - we let it slide
+		cfg->ALPN_PROTOS_LV_LEN = l;
+	} else {
+		if (j - i > 256) {
+			free(cfg->ALPN_PROTOS_LV);
+			cfg->ALPN_PROTOS_LV = NULL;
+			*error = "alpn protocol too long";
+			return (0);
+		}
+		cfg->ALPN_PROTOS_LV[i] = (unsigned char)(j - i - 1);
+		cfg->ALPN_PROTOS_LV_LEN = l + 1;
+	}
+	return (1); // ok!
+}
